/******************************************************************************

	romcheck.c

	Calculates validation & identification checksums for HP9845 ROM dumps
	
	(c) 2006 by Ansgar Kueckes (kueckes@alice-dsl.net)

******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

void print_usage(char *command)
{
	printf("Usage: %s [-plain] [-blocksize <blocksize>] <ROM image>\n", command);
	printf("       <ROM image>  HP combined ROM image file (16K)\n");
	return;
}

/* simulate ADA/ADB machine instruction */

void bpc_unsigned_add(unsigned long *reg, unsigned long summand, int *extend, int *overflow)
{
  long tmp,a,b;
  unsigned long result;

  /* check for overflow condition */  
  a=*reg; if (a>32767) a-=65536;
  b=summand; if (b>32767) b-=65536;
  tmp=a+b;
  if ((tmp<-32768) || (tmp>32767)) *overflow=1;

  /* integer add */
  result=(*reg)+summand;
  
  /* carry from bit 15 ? */
  if (result>65535) *extend=1;
  	
  *reg=result&0xFFFF;
}

void bpc_signed_add(long *reg, long summand, int *extend, int *overflow)
{
  long tmp,a,b,result;

  /* check for overflow condition */  
  tmp=(*reg)+summand;
  if ((tmp<-32768) || (tmp>32767)) *overflow=1;

  /* integer add */
  a=*reg; if (a<0) a+=65536; 
  b=summand; if (b<0) b+=65536;
  result=a+b;
  
  /* carry from bit 15 ? */
  if (result>65535) { result-=65536; *extend=1; }

  if (result>32767) result-=65536;
  *reg=result;
}

void bpc_rar(long *reg, int count)
{
  long a;
  int i,lsb;

  /* integer rotate right */
  a=*reg; if (a<0) a+=65536;
  for (i=0; i<count; i++) {
  	lsb=a&0x01;
  	a>>=1;
  	if (lsb) a|=0x8000;
  }

  if (a>32767) a-=65536;
  *reg=a;
}

void bpc_clear_bit15(long *reg)
{
  long a;
   
  a=*reg; if (a<0) a+=65536;
  a&=0x7FFF;
   
  //if (a>32767) a-=65536;
  *reg=a; 
}

int main(int argc, char *argv[])
{
  int count;
  char *infile = 0;
  FILE *ifp;
  int blocksize=4096;
  unsigned char inbuffer[8192];
  size_t bytes_read;
  long checksum;
  long verify;
  int overflow, extend, excount, ovcount;
  long z;
  int i,j;
  int use_advanced_checksum = 1;

  if (argc < 2) { print_usage(argv[0]); exit(0); }

  /* command line processing */
  for ( count = 1; count < argc ; count++)
  {
  	if (!strcmp(argv[count], "-plain")) {
  		use_advanced_checksum = 0;
  		count++;
  	}
  		
  	if (!strcmp(argv[count], "-blocksize")) {
  		count++; blocksize = strtoul(argv[count],0,0);
  		count++;
  	}

	if (!strncmp(argv[count], "-", 1) && (count < argc))
		 { print_usage(argv[0]); exit(0); }

	if (!infile) { infile = argv[count]; continue; }
  }
  if (!infile) { print_usage(argv[0]); exit(0); }
  
  /* open ROM image files */
  ifp = fopen( infile, "rb");
  if (!ifp) { fprintf(stderr, "Could not open file \"%s\".\n", infile); exit(-1); }
  
  i=0;
  while (1) {
    checksum=0;
    verify=0;
    overflow=0;
    excount=0;
    ovcount=0;
  	
	/* read in one single ROM block */
	if ((bytes_read = fread((void *)inbuffer, 1, blocksize*2, ifp)) != blocksize*2) {
			exit(-1);
	}

	for (j=0; j<blocksize; j++) {
	  	z=inbuffer[j*2]*256+inbuffer[j*2+1];
	  	if (z>32767) z-=65536;
	  	
	  	/* update verify */ 
		extend=0;
		bpc_signed_add(&verify, z, &extend, &overflow);

		if (use_advanced_checksum) {
	        while (extend) {
	        	extend=0;
	        	bpc_signed_add(&verify, 1, &extend, &overflow);
	        	excount++;
	        }
    	}
	        	
	  	/* update checksum */        	
        bpc_rar(&checksum, 3);
        if (checksum<0) {
        	//bpc_clear_bit15(&checksum);
        	bpc_rar(&checksum, 2);
        }
		bpc_signed_add(&checksum, z, &extend, &overflow);
        while (extend) {
        	extend=0;
        	bpc_signed_add(&checksum, 1, &extend, &overflow);
        	excount++;
        }
	}
	if (use_advanced_checksum) {
		printf("block %d: verify=%ld checksum=%ld", i, verify, checksum);
		if (verify == -1) printf ("\t(ROM OK)\n"); else printf ("\t(ROM currupt)\n");
	}
	else {
		printf("block %d: verify=%ld checksum=%ld", i, verify, checksum);
		if (verify == -1) printf ("\t(ROM OK)\n");
		else if (verify == 0) printf ("\t(no ROM present)\n");
		else printf ("\t(ROM currupt)\n");
	}
	
	i++;
  }
  
  /* cleanup */
  if (ifp) fclose(ifp);
  exit(0);
}
